/* SPDX-License-Identifier: GPL-2.0 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <linux/if_link.h>
#include <linux/if_bridge.h>
#include <linux/if_ether.h>
#include <string.h>

#include "json_print.h"
#include "libnetlink.h"
#include "br_common.h"
#include "utils.h"

static unsigned int filter_index;

static void usage(void)
{
	fprintf(stderr,
		"Usage: bridge vni { add | del } vni VNI_ID group IP_ADDRESS dev DEV\n"
		"       bridge vni { show }\n");
	exit(-1);
}

static int parse_vni_filter(const char *argv, struct nlmsghdr *n, int reqsize,
			    __u16 bridge_flags, inet_prefix *group)
{
	struct rtattr *afspec, *nlvlist, *nlvlist_e;
	char *vnilist = strdupa(argv);
	char *vni = strtok(vnilist, ",");
	int group_type = AF_UNSPEC;
	char *v;
	int i;

	afspec = addattr_nest(n, 1024, IFLA_AF_SPEC);

	if (bridge_flags)
		addattr16(n, reqsize, IFLA_BRIDGE_FLAGS, bridge_flags);

	nlvlist = addattr_nest(n, 1024, IFLA_VXLAN_VNI_FILTER | NLA_F_NESTED);

	if (group && is_addrtype_inet(group))
		group_type = (group->family == AF_INET) ?  IFLA_VXLAN_VNI_ENTRY_GROUP :
						     IFLA_VXLAN_VNI_ENTRY_GROUP6;

	for (i = 0; vni; i++) {
		__u32 vni_start = 0, vni_end = 0;

		v = strchr(vni, '-');
		if (v) {
			*v = '\0';
			v++;
			vni_start = atoi(vni);
			vni_end = atoi(v);
		} else {
			vni_start = atoi(vni);
		}
		nlvlist_e = addattr_nest(n, 1024, IFLA_VXLAN_VNI_ENTRY);
		addattr32(n, 1024, IFLA_VXLAN_VNI_ENTRY_START, vni_start);
		if (vni_end)
			addattr32(n, 1024, IFLA_VXLAN_VNI_ENTRY_END, vni_end);
		if (group)
			addattr_l(n, 1024, group_type, group->data, group->bytelen);
		addattr_nest_end(n, nlvlist_e);
		vni = strtok(NULL, ",");
	}
	addattr_nest_end(n, nlvlist);
	addattr_nest_end(n, afspec);

	return 0;
}

static int vni_modify(int cmd, int argc, char **argv)
{
	struct {
		struct nlmsghdr	n;
		struct ifinfomsg	ifm;
		char			buf[1024];
	} req = {
		.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg)),
		.n.nlmsg_flags = NLM_F_REQUEST,
		.n.nlmsg_type = cmd,
		.ifm.ifi_family = PF_BRIDGE,
	};
	unsigned short bridge_flags = BRIDGE_FLAGS_SELF;
	bool group_present = false;
	inet_prefix daddr;
	char *d = NULL;
	char *vni = NULL;

	while (argc > 0) {
		if (strcmp(*argv, "dev") == 0) {
			NEXT_ARG();
			d = *argv;
		} else if (strcmp(*argv, "vni") == 0) {
			NEXT_ARG();
			if (vni)
				invarg("duplicate vni", *argv);
			vni = *argv;
		} else if (!matches(*argv, "group")) {
			NEXT_ARG();
			if (group_present)
				invarg("duplicate group", *argv);
			get_addr(&daddr, *argv, AF_INET);
			if (!is_addrtype_inet_multi(&daddr))
				invarg("invalid group address", *argv);
			group_present = true;
		} else {
			if (matches(*argv, "help") == 0)
				NEXT_ARG();
		}
		argc--; argv++;
	}

	if (d == NULL || vni == NULL) {
		fprintf(stderr, "Device and VNI ID are required arguments.\n");
		return -1;
	}

	if (!vni && group_present) {
		fprintf(stderr, "Group can only be specified with a vni\n");
		return -1;
	}

	if (vni)
		parse_vni_filter(vni, &req.n, sizeof(req), bridge_flags,
				 (group_present ? &daddr : NULL));

	req.ifm.ifi_index = ll_name_to_index(d);
	if (req.ifm.ifi_index == 0) {
		fprintf(stderr, "Cannot find vxlan device \"%s\"\n", d);
		return -1;
	}

	if (rtnl_talk(&rth, &req.n, NULL) < 0)
		return -1;

	return 0;
}

static void open_vni_port(int ifi_index, const char *fmt)
{
	open_json_object(NULL);
	print_color_string(PRINT_ANY, COLOR_IFNAME, "ifname", fmt,
			   ll_index_to_name(ifi_index));
	open_json_array(PRINT_JSON, "vnis");
}

static void close_vni_port(void)
{
	close_json_array(PRINT_JSON, NULL);
	close_json_object();
}

static void print_range(const char *name, __u32 start, __u32 id)
{
	char end[64];

	snprintf(end, sizeof(end), "%sEnd", name);

	print_uint(PRINT_ANY, name, "\t %u", start);
	if (start != id)
		print_uint(PRINT_ANY, end, "-%u", id);

}

static void print_vni(struct rtattr *tb, int ifindex)
{
	struct rtattr *i, *list = tb;
	int rem = RTA_PAYLOAD(list);

	for (i = RTA_DATA(list); RTA_OK(i, rem); i = RTA_NEXT(i, rem)) {
		struct rtattr *ttb[IFLA_VXLAN_VNI_ENTRY_MAX+1];
		__u32 vni_start = 0;
		__u32 vni_end = 0;

		if (i->rta_type != IFLA_VXLAN_VNI_ENTRY)
			continue;

		parse_rtattr(ttb, IFLA_VXLAN_VNI_ENTRY_MAX, RTA_DATA(i),
			     RTA_PAYLOAD(i));

		if (ttb[IFLA_VXLAN_VNI_ENTRY_START])
			vni_start = rta_getattr_u32(ttb[IFLA_VXLAN_VNI_ENTRY_START]);

		if (ttb[IFLA_VXLAN_VNI_ENTRY_END])
			vni_end = rta_getattr_u32(ttb[IFLA_VXLAN_VNI_ENTRY_END]);

		open_json_object(NULL);
		if (vni_end)
			print_range("vni", vni_start, vni_end);
		else
			print_uint(PRINT_ANY, "vni", " %u ", vni_start);
		if (ttb[IFLA_VXLAN_VNI_ENTRY_GROUP]) {
			__be32 addr = rta_getattr_u32(ttb[IFLA_VXLAN_VNI_ENTRY_GROUP]);

			if (addr) {
				if (IN_MULTICAST(ntohl(addr)))
					print_string(PRINT_ANY,
						     "group",
						     "group %s ",
						     format_host(AF_INET, 4, &addr));
				else
					print_string(PRINT_ANY,
						     "remote",
						     "remote %s ",
						     format_host(AF_INET, 4, &addr));
			}
		} else if (ttb[IFLA_VXLAN_VNI_ENTRY_GROUP6]) {
			struct in6_addr addr;

			memcpy(&addr, RTA_DATA(ttb[IFLA_VXLAN_VNI_ENTRY_GROUP6]), sizeof(struct in6_addr));
			if (!IN6_IS_ADDR_UNSPECIFIED(&addr)) {
				if (IN6_IS_ADDR_MULTICAST(&addr))
					print_string(PRINT_ANY,
						     "group6",
						     "group %s ",
						     format_host(AF_INET6,
								 sizeof(struct in6_addr),
								 &addr));
				else
					print_string(PRINT_ANY,
						     "remote6",
						     "remote %s ",
						     format_host(AF_INET6,
								 sizeof(struct in6_addr),
								 &addr));
			}
		}
		close_json_object();
		print_string(PRINT_FP, NULL, "%s", _SL_);
	}
}

static void print_vni_filter(struct rtattr *tb, int ifindex)
{
	struct rtattr *i, *list = tb;
	int rem = RTA_PAYLOAD(list);

	open_vni_port(ifindex, "%s");

	for (i = RTA_DATA(list); RTA_OK(i, rem); i = RTA_NEXT(i, rem)) {

		if (i->rta_type != IFLA_BRIDGE_VXLAN_VNI_FILTER)
			continue;

		print_vni(i, ifindex);
	}
	close_vni_port();
}

static int print_vni_info(struct nlmsghdr *n, void *arg)
{
	struct ifinfomsg *ifm = NLMSG_DATA(n);
	int len = n->nlmsg_len;
	struct rtattr *tb[IFLA_MAX+1];

	if (n->nlmsg_type != RTM_NEWLINK) {
		fprintf(stderr, "Not RTM_NEWLINK: %08x %08x %08x\n",
			n->nlmsg_len, n->nlmsg_type, n->nlmsg_flags);
		return 0;
	}

	len -= NLMSG_LENGTH(sizeof(*ifm));
	if (len < 0) {
		fprintf(stderr, "BUG: wrong nlmsg len %d\n", len);
		return -1;
	}

	if (ifm->ifi_family != AF_BRIDGE)
		return 0;

	if (filter_index && filter_index != ifm->ifi_index)
		return 0;

	parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifm), len);

	/* if AF_SPEC isn't there, vni table is not preset for this port */
	if (!tb[IFLA_AF_SPEC]) {
		if (!is_json_context()) {
			color_fprintf(stdout, COLOR_IFNAME, "%s",
				      ll_index_to_name(ifm->ifi_index));
			fprintf(stdout, "\tNone\n");
		}
		return 0;
	}

	print_vni_filter(tb[IFLA_AF_SPEC], ifm->ifi_index);

	print_string(PRINT_FP, NULL, "%s", _SL_);

	fflush(stdout);
	return 0;
}

static int vni_show(int argc, char **argv)
{
	char *filter_dev = NULL;
	int ret = 0;

	while (argc > 0) {
		if (strcmp(*argv, "dev") == 0) {
			NEXT_ARG();
			if (filter_dev)
				duparg("dev", *argv);
			filter_dev = *argv;
		}
		argc--; argv++;
	}

	if (filter_dev) {
		filter_index = ll_name_to_index(filter_dev);
		if (!filter_index)
			return nodev(filter_dev);
	}

	new_json_obj(json);

	if (!show_stats) {
		if (rtnl_linkdump_req_filter(&rth, PF_BRIDGE,
					     RTEXT_FILTER_BRIDGE_VNI_FILTER) < 0) {
			perror("Cannot send dump request");
			exit(1);
		}

		if (!is_json_context()) {
			printf("dev\tvni ids");
			printf("\n");
		}

		ret = rtnl_dump_filter(&rth, print_vni_info, NULL);
		if (ret < 0) {
			fprintf(stderr, "Dump ternminated\n");
			exit(1);
		}
	}

	delete_json_obj();
	fflush(stdout);
	return 0;
}

int do_vni(int argc, char **argv)
{
	ll_init_map(&rth);

	if (argc > 0) {
		if (matches(*argv, "add") == 0)
			return vni_modify(RTM_SETLINK, argc-1, argv+1);
		if (matches(*argv, "delete") == 0)
			return vni_modify(RTM_DELLINK, argc-1, argv+1);
		if (matches(*argv, "show") == 0 ||
		    matches(*argv, "lst") == 0 ||
		    matches(*argv, "list") == 0)
			return vni_show(argc-1, argv+1);
		if (matches(*argv, "help") == 0)
			usage();
	} else {
		return vni_show(0, NULL);
	}

	fprintf(stderr, "Command \"%s\" is unknown, try \"bridge vni help\".\n", *argv);
	exit(-1);
}
